/*======================================================================

    Silicom PCMCIA client driver

    Written by Igor Royzis, igor@silicom.co.il
    Based upon skeleton driver by David Hinds, dhinds@allegro.stanford.edu

    Revision 1.0 Beta 1999/1/28
======================================================================*/

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <asm/system.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/ioport.h>
#include <linux/wanpipe.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>
#include "silspe.h"
#include "speloc.h"


char *version = "spe_cs.c 1.02 1999/1/25 (Silicom)";


/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* Bit map of interrupts to choose from */
/* This means pick from 15, 14, 12, 11, 10, 9, 7, 5, 4, and 3 */
static u_int irq_mask = 0xdeb8;

MODULE_PARM(irq_mask, "i");

/*====================================================================*/

/*
   The event() function is this driver's Card Services event handler.
   It will be called by Card Services when an appropriate card status
   event is received.  The config() and release() entry points are
   used to configure or release a socket, in response to card insertion
   and ejection events.  They are invoked from the skeleton event
   handler.
*/

static int spe_config(struct net_device *dev, struct ifmap *map);
static void spe_release(u_long arg);
static int  spe_init(struct net_device *dev);
static int spe_event(event_t event, int priority, event_callback_args_t *args);

/*
   The attach() and detach() entry points are used to create and destroy
   "instances" of the driver, where each instance represents everything
   needed to manage one actual PCMCIA card.
*/

static dev_link_t *spe_attach(void);
static void spe_detach(dev_link_t *);


/*
   The dev_info variable is the "key" that is used to match up this
   net_device driver with appropriate cards, through the card configuration
   database.
*/

static dev_info_t dev_info = "spe_cs";

/*
   A linked list of "instances" of the SPE net_device.  Each actual
   PCMCIA card corresponds to one net_device instance, and is described
   by one dev_link_t structure (defined in ds.h).

   You may not want to use a linked list for this -- for example, the
   memory card driver uses an array of dev_link_t pointers, where minor
   net_device numbers are used to derive the corresponding array index.
*/

static dev_link_t *dev_list = NULL;



/*====================================================================*/

static void cs_error(client_handle_t handle, int func, int ret)
{
error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}

/*======================================================================
    spe_attach() creates an "instance" of the driver, allocating
    local data structures for one net_device.  The net_device is registered
    with Card Services.

    The dev_link structure is initialized, but we don't actually
    configure the card at this point -- we wait until we receive a
    card insertion event.

    The function returns a pointer to new dev_link_t structure or NULL
    if new instance could not be created.
======================================================================*/
static dev_link_t *spe_attach(void)
{
client_reg_t client_reg;
dev_link_t *link;
struct net_device *dev;
int ret;


    DEBUG(5, "spe_attach()\n")

    /* Initialize the dev_link_t structure */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    memset(link, 0, sizeof(struct dev_link_t));

    link->release.function = &spe_release;
    link->release.data = (u_long)link;

    /* The io structure describes IO port mapping */
    link->io.NumPorts1 = 16;
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
    link->io.IOAddrLines = 4;

    /* Interrupt setup */
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE | IRQ_HANDLE_PRESENT;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
    link->irq.IRQInfo2 = irq_mask;
    link->irq.Handler = spe_interrupt;

    /* General socket configuration */
    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;
    link->conf.ConfigIndex = 1;
    link->conf.Present = PRESENT_OPTION;

    /* Allocate space for SPE net_device-specific data */
    dev = kmalloc(sizeof(struct net_device), GFP_KERNEL);
    memset(dev, 0, sizeof(struct net_device));

    /* Make up a SPE-specific-data structure. */
    dev->priv = kmalloc(sizeof(struct spe_local), GFP_KERNEL);
    memset(dev->priv, 0, sizeof(struct spe_local));

    /* The SPE-specific entries in the net_device structure. */
    dev->hard_start_xmit = &spe_start_xmit;
    dev->set_config = &spe_config;
    dev->get_stats = &spe_get_stats;
    dev->set_multicast_list = &set_multicast_list;
    //dev->tbusy = 1;
    stop_net_queue(dev);
    ether_setup(dev);
    //dev->name = ((struct spe_local*)dev->priv)->node.dev_name;
    init_dev_name(dev, ((struct spe_local*)dev->priv)->node);
    dev->init = &spe_init;
    dev->open = &spe_open;
    dev->stop = &spe_close;
    link->priv = link->irq.Instance = dev;

    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
	CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
	CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
	CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &spe_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != CS_SUCCESS) {
	cs_error(link->handle, RegisterClient, ret);
        spe_detach(link);
	return NULL;
    }

    return link;
} /* spe_attach */

/*======================================================================
    This deletes a driver "instance".  The net_device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the net_device is released.
======================================================================*/
static void spe_detach(dev_link_t *link)
{
dev_link_t **linkp;

    DEBUG(5, "spe_detach(0x%p)\n", link)

    /* Locate net_device structure in our list */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
	if (*linkp == link) break;
    if (*linkp == NULL)
	return;

    /* If the net_device is currently configured and active, release it */
    if (link->state & DEV_CONFIG)
        spe_release((u_long)link);

    /* Break the link with Card Services */
    if (link->handle)
	CardServices(DeregisterClient, link->handle);

    /* Unlink net_device structure, free pieces */
    *linkp = link->next;
    if (link->priv) {
       struct net_device *dev = link->priv;
       if(dev->priv)
          kfree_s(dev->priv, sizeof(struct spe_local));
       kfree_s(link->priv, sizeof(struct net_device));
    }
    kfree_s(link, sizeof(struct dev_link_t));

} /* spe_detach */

/*======================================================================
    spe_requestconfig() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    ethernet net_device available to the system.
======================================================================*/

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn),args))!=0) goto cs_failed

static void spe_requestconfig(dev_link_t *link)
{
client_handle_t handle;
tuple_t tuple;
cisparse_t parse;
struct net_device *dev;
int i, j, last_fn, last_ret;
u_char buf[64];
i = 0;
handle = link->handle;
dev = link->priv;

    DEBUG(1, "spe_config(0x%p)\n", link)

    /*
       This reads the card's CONFIG tuple to find its configuration
       registers.
    */
    tuple.DesiredTuple = CISTPL_CONFIG;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    tuple.TupleData = buf;
    tuple.TupleDataMax = 64;
    tuple.TupleOffset = 0;
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    link->conf.ConfigBase = parse.config.base;


    /* Configure card */
    link->state |= DEV_CONFIG;

    /*
	Allocating IO ports. 
    */
    for (j = 0; j < 0x400; j += 0x10) {
	/* The '^0x300' is so that we probe 0x300-0x3ff first, then
	   0x200-0x2ff, and so on, because this seems safer */
      link->io.BasePort1 = j^0x300;
      i = CardServices(RequestIO, link->handle, &link->io);
      if (i == CS_SUCCESS) break;
	}
    if (i != CS_SUCCESS) {
      cs_error(link->handle, RequestIO, i);
      goto failed;
    }

    /*
       Now allocate an interrupt line.  Note that this does not
       actually assign a handler to the interrupt.
    */
    CS_CHECK(RequestIRQ, link->handle, &link->irq);

    /*
       This actually configures the PCMCIA socket -- setting up
       the I/O windows and the interrupt mapping.
    */
    CS_CHECK(RequestConfiguration, link->handle, &link->conf);


    /* Initialize the net_device */
    dev->irq = link->irq.AssignedIRQ;
    dev->base_addr = link->io.BasePort1;
    /* read node address dev->dev_addr */
    tuple.DesiredTuple = CISTPL_FUNCE;
    tuple.Attributes = TUPLE_RETURN_COMMON;
    if ((CardServices(GetFirstTuple, handle, &tuple) == CS_SUCCESS) &&
	(CardServices(GetTupleData, handle, &tuple) == CS_SUCCESS)) {
       char *tmp = &buf[6];            /* Skip junk in the tuple */
       if(*tmp == 6)  {                /* Byte at offset 6 must be 6 */
         for(i=0,tmp++; i < 6; i++, tmp++)
           dev->dev_addr[i] = *tmp;
       }
       else
        DEBUG(1, "spe_cs: cant read node address, wrong tuple format\n")
    }
    else
        DEBUG(1, "spe_cs: cant read node address, no tuple found\n")

    //dev->tbusy = 0;
    start_net_queue(dev);	
    if(register_netdev(dev) != 0)  {
        DEBUG(1, "spe_cs: register_netdev() failed\n")
	goto failed;
    }


    link->state &= ~DEV_CONFIG_PENDING;
    link->dev = &((struct spe_local*)dev->priv)->node;
    copy_dev_name(((struct spe_local*)dev->priv)->node, dev);

    DEBUG(1, "spe net_device loaded\n")

    // Initialize the chip (lower level init)
    if(spe_drvinit(dev) != 0)  {
       DEBUG(1, "spe chip initialization failed\n")
       goto failed;
    }
    return;

cs_failed:
    cs_error(link->handle, last_fn, last_ret);
failed:
    spe_release((u_long)link);

} /* spe_requestconfig */

/*======================================================================

    After a card is removed, spe_release() will unregister the net
    net_device, and release the PCMCIA configuration.  If the net_device is
    still open, this will be postponed until it is closed.

======================================================================*/

static void spe_release(u_long arg)
{
dev_link_t *link = (dev_link_t *)arg;
struct net_device *dev = link->priv;

    DEBUG(5, "spe_release(0x%p)\n", link)

    /*
       If the net_device is currently in use, we won't release until it
       is actually closed.
    */
    if (link->open) {
        DEBUG(1, "spe_cs: release postponed, '%s' still open\n",
	         link->dev->dev_name)
	link->state |= DEV_STALE_CONFIG;
	return;
    }


    /* Release the resources */
    CardServices(ReleaseConfiguration, link->handle);
    CardServices(ReleaseIO, link->handle, &link->io);
    CardServices(ReleaseIRQ, link->handle, &link->irq);

    if(link->dev)
      unregister_netdev(dev);

    link->dev = NULL;
    link->state &= ~DEV_CONFIG;

    if (link->state & DEV_STALE_LINK)
       spe_detach(link);

} /* spe_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the net drivers from trying
    to talk to the card any more.

    When a CARD_REMOVAL event is received, we immediately set a flag
    to block future accesses to this net_device.  All the functions that
    actually access the net_device should check this flag to make sure
    the card is still present.

======================================================================*/

static int spe_event(event_t event, int priority,
			  event_callback_args_t *args)
{
dev_link_t *link = args->client_data;
struct net_device *dev = link->priv;
struct spe_local *local = dev->priv;

    DEBUG(5, "spe_event(0x%06x)\n", event)

    switch (event) {
    case CS_EVENT_CARD_REMOVAL:
	link->state &= ~DEV_PRESENT;
	spe_close(dev);
	if (link->state & DEV_CONFIG) {
	    //dev->tbusy = 1; 
	    //stop_net_queue(dev);
            netif_device_detach(dev);
	    local->start = 0;
	    //dev->start = 0;
	    //link->release.expires = RUN_AT(HZ/20);
	    //add_timer(&link->release);
	    mod_timer(&link->release, jiffies + HZ/20);
	}
	break;
    case CS_EVENT_CARD_INSERTION:
	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
        spe_requestconfig(link);
	break;
    case CS_EVENT_PM_SUSPEND:
	link->state |= DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
	if (link->state & DEV_CONFIG)  {
            if(link->open)  {
               //dev->tbusy=1; 
		//stop_net_queue(dev);
		//dev->start=0;
		local->start=0;
		netif_device_detach(dev);
            }
	    CardServices(ReleaseConfiguration, link->handle);
        }
	break;
    case CS_EVENT_PM_RESUME:
	link->state &= ~DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_CARD_RESET:
	if (link->state & DEV_CONFIG)  {
	    CardServices(RequestConfiguration, link->handle, &link->conf);
	    dev->stop = 0;
            if(link->open)  {
              spe_reset(dev);
              //dev->tbusy=0;
	      //start_net_queue(dev); 
	      //dev->start=1;
	      local->start = 1;
	      netif_device_attach(dev);
            }
        }
	break;
    }
    return 0;
} /* spe_event */

/*
 Nothing to do in initialization since we work only with already installed
 cards
*/
static int spe_config(struct net_device *dev, struct ifmap *map)
{
    return 0;
}

static int  spe_init(struct net_device *dev)
{
    return 0;
}

/*====================================================================
 init_module is called upon when the module is installed.
 It registers the driver with Device Services
=====================================================================*/

int init_module(void)
{
    servinfo_t serv;
    DEBUG(5, "%s\n", version)
    CardServices(GetCardServicesInfo, &serv);
    if (serv.Revision != CS_RELEASE_CODE) {
	printk(KERN_NOTICE "spe: Card Services release does not match!\n");
	return -1;
    }
    register_pcmcia_driver(&dev_info, &spe_attach, &spe_detach);
    return 0;
}

/*====================================================================
 cleanup_module is called upon when the module is removed.
 It deregisters the driver with Device Services and cleanups the
 allocated memory
=====================================================================*/
void cleanup_module(void)
{
    DEBUG(5, "spe_cs: unloading\n")
    unregister_pcmcia_driver(&dev_info);
    /*
     Detach all the instances and cleanup the memory
    */
    while (dev_list != NULL) {
        spe_detach(dev_list);
    }
}

